#include "LineInterpolator2D.h"
#include <cstdlib>


bool
CurrentXSmallerCurrentYSmaller::
operator()(const LineInterpolator2D &l1, const LineInterpolator2D &l2) const {
  // Decide by x component if possible
  int x1 = l1.GetX();
  int x2 = l2.GetX();
  if (x1 != x2) return x1 < x2;
  
  // Decide by y component if possible
  int y1 = l1.GetY();
  int y2 = l2.GetY();
  if (y1 != y2) return y1 < y2;

  // Last measure: Decide by error term
  int e1 = l1.ErrorTerm();
  int e2 = l2.ErrorTerm();
  return (e1 < e2);
}

bool
CurrentXSmallerCurrentYSmaller::
operator()(const LineInterpolator2D *l1, const LineInterpolator2D *l2) const {
  return this->operator()(*l1, *l2);
}





LineInterpolator2D::
LineInterpolator2D(int x1, int y1, int x2, int y2) :
  m_x1(x1), m_y1(y1), m_x2(x2), m_y2(y2) {
  // Initialize bresenham-like interpolation of the y coordinate. The
  // current integer y is floor(y').
  int dx = m_x2 - m_x1;
  int adx = std::abs(dx);
  int dy = m_y2 - m_y1;

  m_xinc = (dx < 0 ? -1 : 1);
 
  // we need m_yinc = floor(dy/adx). The following is an integer only way to
  // say that.
  if (adx == 0) {
    m_yinc = 0;
  } else {
    m_yinc = dy / adx;
    if (dy < 0 && dy % adx != 0)
      --m_yinc;
  }

  m_yerrorinc = dy - m_yinc * adx;
  m_ymaxerror = adx;
  m_yoverflow = 1;
  m_yreduceerror = m_ymaxerror;

  GoToBegin();
}

void
LineInterpolator2D::
GoToBegin() {
  m_x = m_x1;
  m_y = m_y1;
  m_yerror = 0;
  m_stepsleft = std::abs(m_x2 - m_x1);
}

LineInterpolator2D&
LineInterpolator2D::
operator++() {
  // Take a unit step in x direction, update y to floor(y') for the floating
  // point y' corresponding to the new x on the line.
  //
  // NO BOUNDS CHECKING!
  m_x += m_xinc;
  m_y += m_yinc;
  m_yerror += m_yerrorinc;
  if (m_yerror >= m_ymaxerror) {
    m_y += m_yoverflow;
    m_yerror -= m_yreduceerror;
  }
  --m_stepsleft;

  return *this;
}

int
LineInterpolator2D::
GetX() const {
  return m_x;
}
  
int
LineInterpolator2D::
GetY() const {
  return m_y;
}

bool
LineInterpolator2D::
IsYExact() const {
  return !m_yerror;
}

bool
LineInterpolator2D::
IsAtEnd() const {
  return m_stepsleft < 0;
}

bool
LineInterpolator2D::
IsAtLastElement() const {
  return m_stepsleft == 0;
}

int
LineInterpolator2D::
ErrorTerm() const {
  return m_yerror;
}
