#include "PolygonScanConverter.h"

PolygonScanConverter::
PolygonScanConverter(VertexListType &vertices) {
  // Constructor
  //
  // The list of vertices v1, ..., vn is first converted into a list of edges by
  // pairing adjacent vertices. (vn, v1) is an edge, too.
  //
  // Tracing of an edge starts at the smaller x coordinate.
  //
  // Edges with constant x are not traced at all.

  // Construct a list of all edges
  for (VertexListType::iterator it = vertices.begin(); it != vertices.end(); ++it) {
    const VertexType &v1 = *it;
    const VertexType &v2 = (it+1 == vertices.end() ? *(vertices.begin()) : *(it+1));

    // Skip vertical lines
    if (v1.x == v2.x)
      continue;

    // Scan from low x to high x
    LineInterpolator2D *line;
    if (v1.x < v2.x) {
      line = new LineInterpolator2D(v1.x, v1.y, v2.x, v2.y);
    } else {
      line = new LineInterpolator2D(v2.x, v2.y, v1.x, v1.y);
    }

    m_globalEdgeList.insert(line);
  }

  // Initialize the iterator functionality
  GoToBegin();
}

PolygonScanConverter::
~PolygonScanConverter() {
  // Destructor deletes all the LineInterpolator2D's
  for (EdgeListType::const_iterator it = m_globalEdgeList.begin();
      it != m_globalEdgeList.end();
      ++it)
    delete *it;

  for (EdgeListType::const_iterator it = m_activeEdgeList.begin();
      it != m_activeEdgeList.end();
      ++it)
    delete *it;

  for (EdgeListType::const_iterator it = m_finishedEdgeList.begin();
      it != m_finishedEdgeList.end();
      ++it)
    delete *it;
}

void 
PolygonScanConverter::
GoToBegin() {
  // Method to set the scan conversion iterator to the first pixel
  //
  // Moves all lines from the active and finished egde lists to the global
  // edge list. At the same time, resets the line iterators
  for (EdgeListType::const_iterator it = m_activeEdgeList.begin();
      it != m_activeEdgeList.end(); ++it) {
    LineInterpolator2D *line = *it;
    line->GoToBegin();
    m_globalEdgeList.insert(line);
  }
  m_activeEdgeList.clear();

  for (EdgeListType::const_iterator it = m_finishedEdgeList.begin();
      it != m_finishedEdgeList.end(); ++it) {
    LineInterpolator2D *line = *it;
    line->GoToBegin();
    m_globalEdgeList.insert(line);
  }
  m_finishedEdgeList.clear();

  // Without any edges there are no points inside the polygon
  if (m_globalEdgeList.empty()) {
    m_finished = true;
    return;
  }

  // Initialize the active edge list. If this returns false, there are no
  // active edges, the polygon is empty, and we are done.
  m_finished = false;
  if (!AdvanceActiveEdges())
    return;

  // Now find the line where the filling starts
  if (!GoToFirstScanRowSegment()) {
    GoToNextScanRow();
  }
}

int 
PolygonScanConverter::
GetX() const {
  return m_currentX;
}

int 
PolygonScanConverter::
GetY() const {
  return m_currentY;
}

bool 
PolygonScanConverter::
IsAtEnd() const {
  return m_finished;
}

int 
PolygonScanConverter::
NumberOfPixelsToEndOfSegment(EdgeListType::const_iterator currentEdge, int y) {
  // currentEdge and currentEdge+1 make a scan row segment from ymin to
  // ymax. NumberOfPixelsToEndOfSegment returns the number of pixels from
  // y to ymax excluding y.
  EdgeListType::const_iterator endEdge = currentEdge;
  ++endEdge;

  int final = (*endEdge)->GetY();
  int left = final - y;

  return (left > 0 ? left : 0);
}

bool 
PolygonScanConverter::
GoToFirstNonEmptyScanRowSegmentAtOrAfter(EdgeListType::const_iterator oddActiveEdge) {
  // Method to position the iterator at the beginning of the scan row segment
  // at or below the given active edge. The segment has to contain at least
  // one pixel.
  //
  // oddActiveEdge must be just that!
  //
  // If true is returned, the iterator is at it's new position.
  // If false is returned, there is no segment to go to in the current scan
  // row and the iterator is left where it was.
  EdgeListType::const_iterator newCurrentEdge = oddActiveEdge;

  EdgeListType::const_iterator endEdge = newCurrentEdge;
  ++endEdge;

  while (newCurrentEdge != m_activeEdgeList.end()) {
    LineInterpolator2D *line = *newCurrentEdge;

    int startY = line->GetY();
    // startY is either on the polygon edge or just outside of it.
    // Find the first pixel that might be inside the polygon
    if (!line->IsYExact())
      ++startY;

    // Because of the increase, the segment might be empty.
    // NumberOfPixelsToEndOfSegment returns the number excluding the start
    // pixel, so we subtract 1 from startY
    if (NumberOfPixelsToEndOfSegment(newCurrentEdge, startY-1) == 0) {
      ++newCurrentEdge;
      ++newCurrentEdge; // Odd/Even rule for polygon filling
    } else {
      // A good segment was found! Now let's move that iterator!
      m_currentEdge = newCurrentEdge;
      m_currentX = line->GetX();
      m_currentY = startY;
      return true;
    }
  }

  // No segment could be found
  return false;
}

bool 
PolygonScanConverter::
GoToFirstScanRowSegment() {
  return GoToFirstNonEmptyScanRowSegmentAtOrAfter(m_activeEdgeList.begin());
}

bool 
PolygonScanConverter::
GoToNextScanRowSegment() {
  EdgeListType::const_iterator nextCurrentEdge = m_currentEdge;
  ++nextCurrentEdge;
  ++nextCurrentEdge;

  if (nextCurrentEdge == m_activeEdgeList.end())
    return false;
  else
    return GoToFirstNonEmptyScanRowSegmentAtOrAfter(nextCurrentEdge);
}

bool 
PolygonScanConverter::
AdvanceActiveEdges() {
  // Edges that reached their endpoint are moved to the list of finished
  // edges. If global edges start at the new scan row, they are moved to the
  // active edges.
  //
  // False is returned if there is no new scan row. In this case, m_finished
  // is set to true as well.
  //
  // The iterator position is not changed.

  // Update the active edges and move them to a temporary storage to keep
  // them sorted after the update.
  EdgeListType temporaryActiveEdgeList;
  EdgeListType::const_iterator it = m_activeEdgeList.begin();
  while (it != m_activeEdgeList.end()) {
    LineInterpolator2D *line = *it;

    // Remove the line from the list of active edges. If it is still active
    // after it has been updated, it will be inserted again.
    EdgeListType::const_iterator deleteIt = it;
    ++it;
    m_activeEdgeList.erase(deleteIt);

    if (line->IsAtLastElement()) {
      m_finishedEdgeList.insert(line);
      continue;
    }

    // Advance active edge. If there are still global edges left, one of them
    // will start at the end of the current edge. In this case, we don't visit
    // the last pixel of the current edge.
    ++(*line);
    if (!m_globalEdgeList.empty() && line->IsAtLastElement()) {
      m_finishedEdgeList.insert(line);
    } else {
      temporaryActiveEdgeList.insert(line);
    }
  }
  // copy the temporary edge list back in the active edge list
  m_activeEdgeList = temporaryActiveEdgeList;

  // If there are no more active edges, we have to insert the next remaining
  // global edge.
  if (m_activeEdgeList.empty()) {
    // If there are no more global edges, we are done.
    if (m_globalEdgeList.empty()) {
      m_finished = true;
      return false;
    } else {
      // Activate the next global egde
      LineInterpolator2D *line = *(m_globalEdgeList.begin());

      m_activeEdgeList.insert(line);
      m_globalEdgeList.erase(m_globalEdgeList.begin());
    }
  }

  // Now we have at least one active edge. Add global edges that start in
  // the same scan row.
  int currentX = (*m_activeEdgeList.begin())->GetX();
  while (!m_globalEdgeList.empty()) {
    LineInterpolator2D *line = *(m_globalEdgeList.begin());

    if (line->GetX() == currentX) {
      m_activeEdgeList.insert(line);
      m_globalEdgeList.erase(m_globalEdgeList.begin());
    } else {
      // That's it, the other global edges are still too far away.
      break;
    }
  }

  return true;
}

bool 
PolygonScanConverter::
GoToNextScanRow() {
  // Sets the iterator position to the first pixel inside the polygon in the
  // next scan row that contains such a pixel.
  //
  // Returns false iff no such pixel is found.
  AdvanceActiveEdges();
  if (m_finished)
    return false;

  while (!GoToFirstScanRowSegment()) {
    AdvanceActiveEdges();
    if (m_finished)
      return false;
  }
  return true;
}

PolygonScanConverter& 
PolygonScanConverter::
operator++() {
  // Try to proceed in y direction. If there are no more pixels in this row,
  // try to proceed in x direction to the next scan row. If there
  // are no more scan rows, we are done.
  
  // Is a pixel left in the current scan row segment?
  if (NumberOfPixelsToEndOfSegment(m_currentEdge, m_currentY)) {
    ++m_currentY;
    return *this;
  }

  // Find the next scan row segment with at least one pixel
  if (GoToNextScanRowSegment()) {
    return *this;
  }

  // There is no more pixel in this scan row, so we proceed to the next.
  GoToNextScanRow();

  return *this;
}
