#ifndef POLYGONSCANCONVERTER_H
#define POLYGONSCANCONVERTER_H

#include "LineInterpolator2D.h"
#include <vector>
#include <set>

class PolygonScanConverter {
  // A class that iterates over all points with intger coordinates inside a
  // complex polygon.
  //
  // Points exactly on the border are included.
  //
  // It's possible that the set of iterated points is not connected, for
  // example the triangle (0,0), (0,1), (3,2)
  //
  // The implementation follows the active edge table approach with the
  // odd/even inclusion rule, beautifully described in
  // The Simplicity of Complex Polygons, Michael Abrash, Dr. Dobb's Journal
  // June 1991, v16, n6, p139(6)

public:
  typedef struct {
    int x;
    int y;
  } VertexType;

  typedef std::vector<VertexType> VertexListType;

private:
  typedef std::multiset<LineInterpolator2D*, CurrentXSmallerCurrentYSmaller> EdgeListType;

public:
  // Constructor
  //
  // The list of vertices v1, ..., vn is first converted into a list of edges by
  // pairing adjacent vertices. (vn, v1) is an edge, too.
  //
  // Tracing of an edge starts at the smaller x coordinate.
  //
  // Edges with constant x are not traced at all.
  PolygonScanConverter(VertexListType &vertices);

  ~PolygonScanConverter();


  // Method to set the scan conversion iterator to the first pixel
  //
  // Moves all lines from the active and finished egde lists to the global
  // edge list. At the same time, resets the line iterators
  void GoToBegin();

  // Try to proceed in y direction. If there are no more pixels in this row,
  // try to proceed in x direction to the next scan row. If there
  // are no more scan rows, we are done.
  PolygonScanConverter& operator++();

  int GetX() const;

  int GetY() const;

  bool IsAtEnd() const;

protected:
  EdgeListType m_globalEdgeList;
  EdgeListType m_activeEdgeList;
  EdgeListType m_finishedEdgeList;

  int m_currentX;
  int m_currentY;
  EdgeListType::const_iterator m_currentEdge;
  bool m_finished;

private:
  // We probably won't need copy construction or assignment.
  PolygonScanConverter(const PolygonScanConverter&);
  PolygonScanConverter& operator=(const PolygonScanConverter&);

  int NumberOfPixelsToEndOfSegment(EdgeListType::const_iterator currentEdge, int from);

  // Method to position the iterator at the beginning of the scan row
  // segment at or below the given active edge that contains at least one
  // pixel.
  //
  // oddActiveEdge must be just that!
  //
  // If true is returned, the iterator is at it's new position.
  // If false is returned, there is no segment to go to in the current scan
  // row and the iterator is left where it was.
  bool GoToFirstNonEmptyScanRowSegmentAtOrAfter(EdgeListType::const_iterator oddActiveEdge);

  bool GoToFirstScanRowSegment();

  bool GoToNextScanRowSegment();

  // Edges that reached their endpoint are moved to the list of finished
  // edges. If global edges start at the new scan row, they are moved to the
  // active edges.
  //
  // False is returned if there is no new scan row.
  //
  // The iterator position is not changed.
  bool AdvanceActiveEdges();

  bool GoToNextScanRow();
};
#endif
